// Author: Pete Kirkham
//
// quad store b-tree.
//
#ifndef _B_TREE_H_
#define _B_TREE_H_
#define __USE_FILE_OFFSET64
#include <sys/types.h>
#include <xmmintrin.h>

#include "utils/core_utils.h"

static const size_t block_size(256*16);

enum sort_order
{
    spog = 0x1234,
    spgo = 0x1243,
    sopg = 0x1324,
    sogp = 0x1342,
    sgpo = 0x1423,
    sgop = 0x1432,
    psog = 0x2134,
    psgo = 0x2143,
    posg = 0x2314,
    pogs = 0x2341,
    pgso = 0x2413,
    pgos = 0x2431,
    ospg = 0x3124,
    osgp = 0x3142,
    opsg = 0x3214,
    opgs = 0x3241,
    ogsp = 0x3412,
    ogps = 0x3421,
    gspo = 0x4123,
    gsop = 0x4132,
    gpso = 0x4213,
    gpos = 0x4231,
    gosp = 0x4312,
    gops = 0x4321,
};

template <int term>
inline __m128i set_single_term (unsigned int term_value) ;


template <> inline __m128i set_single_term<1> (unsigned int term_value)
{
    _mm_set_epi32(term_value, 0, 0, 0);
}

template <> inline __m128i set_single_term<2> (unsigned int term_value)
{
    _mm_set_epi32(0, term_value, 0, 0);
}

template <> inline __m128i set_single_term<3> (unsigned int term_value)
{
    _mm_set_epi32(0, 0, term_value, 0);
}

template <> inline __m128i set_single_term<4> (unsigned int term_value)
{
    _mm_set_epi32(0, 0, 0, term_value);
}

template <int term>
inline bool bisection_match      (__m128i value, __m128i* data, size_t& lo_in, size_t& hi_in)
{
//    __m128i value(set_single_term<term>(term_value));
    
    if (_mm_extract_epi16(_mm_cmpgt_epi32(_mm_loadu_si128(data), value), 9-(term*2))) return false;
    if (_mm_extract_epi16(_mm_cmplt_epi32(_mm_loadu_si128(data+hi_in-1), value), 9-(term*2))) return false;

    size_t lo(lo_in);
    size_t hi(hi_in);
    
    if (_mm_extract_epi16(_mm_cmpeq_epi32(_mm_loadu_si128(data + lo), value), 9-(term*2))) 
    {
        lo = 0;
    }
    else
    {
        size_t lo_lo(lo_in);
        size_t lo_hi(hi_in-1);
        
        // scan for start
        while (lo_lo != lo_hi)
        {
            size_t i(lo_lo + ((lo_hi-lo_lo)>>1));
            __m128i in(_mm_loadu_si128(data+i));
            __m128i cmp(_mm_cmplt_epi32(in, value));

            //~ std::cout << '(' << lo_lo << ',' << lo_hi << ')'  << to_i4(in) << " ?= " << to_i4(value) << " -> " << to_i4(cmp) << std::endl;
            
            if (_mm_extract_epi16(cmp, 7))
            {
                if (lo_lo == i) break;
                lo_lo = i;
            }
            else
            {
                lo_hi = i;
            }
        }
        
        lo = lo_hi;
    }

    if (_mm_extract_epi16(_mm_cmpeq_epi32(_mm_loadu_si128(data+hi_in-1), value), 7)) 
    {
        hi = hi_in;
    }
    else
    {
        size_t hi_lo(lo);
        size_t hi_hi(hi_in-1);

        // scan for start
        while (hi_lo != hi_hi)
        {
            size_t i(hi_lo + ((hi_hi-hi_lo)>>1));
            __m128i in(_mm_loadu_si128(data+i));
            __m128i cmp(_mm_cmpgt_epi32(in, value));

            //~ std::cout << '(' << hi_lo << ',' << hi_hi << ',' << i << ')' << to_i4(in) << " ?= " << to_i4(value) << " -> " << to_i4(cmp) << std::endl;
            
            if (_mm_extract_epi16(cmp, 7))
            {
                hi_hi = i;
            }
            else
            {
                if (hi_lo == i) break;
                hi_lo = i;
            }
        }
        
        hi = hi_hi + 1;
    }

    if (lo != hi)
    {
        lo_in = lo;
        hi_in = hi;
        
        return true;
    }
    else
    {
        return false;
    }
}

// a reference to a block
class block_ref
{
    
};

// block_access used to access file blocks.
class block_file
{
public:
    block_file (int file) : _file(file), _block_count(0)
    {
        posix_memalign(reinterpret_cast<void**>(&_data), 16, block_size);
    }
    
    ~block_file ()
    {
        free(_data);
    }
    
    // api for accessing blocks of data
    __m128i*    get_block (unsigned int block_id)
    {
        __m128i*    data(_data);
        
        // search for the query in the tree
        ::lseek(_file, block_id * block_size, SEEK_SET);
        ::read(_file, data, block_size);
        
        return data;
    }
    
    void    commit_block (__m128i* data, unsigned int block_id)
    {
        ::lseek(_file, block_id * block_size, SEEK_SET);
        ::write(_file, data, block_size);
    }
    
    void    release_block (__m128i* data, unsigned int block_id)
    {
    }
    
    unsigned int    create_block ()
    {
        return _block_count++;
    }
    
private:
    __m128i*        _data;
    unsigned int    _block_count;
    int             _file;
};

template <sort_order order, class block_accessor = block_file>
class b_tree 
{
public:
    b_tree (int file) : _block_accessor(file)
    {
    }
    
    ~b_tree ()
    {
    }
    
    __m128i*    lookup (__m128i query, unsigned int num_terms, __m128i* result, size_t max_count)
    {
        return lookup_b(query, num_terms, result, max_count, 0);
    }
    
    void        insert (__m128i quad)
    {
        unsigned int block_added(0);
        
        insert_b(quad, 0, block_added);
    }
    
private:
    __m128i* lookup_b (__m128i query, unsigned int num_terms, __m128i* result, size_t max_count, unsigned int block_id)
    {
        __m128i*        data(_block_accessor.get_block(block_id));
        
        // if first int & 0x100 is zero, then is list of offsets rather than data
        // 0 is code, 1-201 items are quads, 202-252 are offsets (packed).
        unsigned int    type(reinterpret_cast<int*>(data)[0]);
        unsigned int    size(type&0x000000ff);
        
        // quads start after first block
        __m128i*        quads(data + 1);

        // scan for each term in sort order
        size_t scan_lo(0);
        size_t scan_hi(size);
        
        if (bisection_match<(order>>12)&0xf>(query, quads, scan_lo, scan_hi)) 
        {
            if ((num_terms < 2) || bisection_match<(order>>8)&0xf>(query, quads, scan_lo, scan_hi))
            {
                if ((num_terms < 3) || bisection_match<(order>>4)&0xf>(query, quads, scan_lo, scan_hi))
                {
                    if ((num_terms < 4) || bisection_match<order&0xf>(query, quads, scan_lo, scan_hi))
                    {
                        // x100 indicates offsets
                        if (type & 0x100)
                        {
                            // scan each block referenced
                            
                            // blocks referenced start at data + 202 to allow 201 quads, so can insert one and then split so 
                            // less that 200.
                            int*    blocks(reinterpret_cast<int*>(data + 202));
                            
                            for (size_t i(scan_lo); i < scan_hi; ++i) 
                            {
                                __m128i* new_result(lookup_b(query, num_terms, result, max_count, blocks[i]));
                                max_count -= new_result - result;
                                result = new_result;
                            }
                        }
                        else // zero indicates data
                        {
                            // scan data for matches
                            for (__m128i *p(data+1+scan_lo), *q(data+1+scan_hi), *r(result); p < q; ++p, ++r) 
                            {
                                _mm_store_si128(r, _mm_load_si128(p));
                            }
                            
                            result += scan_hi - scan_lo;
                        }
                    }
                }
            }
        }
        
        return result;
    }
    
    
    bool insert_b (__m128i quad, unsigned int block_id, unsigned int& block_added, __m128i& break_value)
    {
        __m128i*        data(_block_accessor.get_block(block_id));
        
        // if first int & 0x100 is zero, then is list of offsets rather than data
        // 0 is code, 1-201 items are quads, 202-252 are offsets (packed).
        unsigned int&   type(reinterpret_cast<int*>(data)[0]);
        unsigned int    size(type&0x000000ff);
        
        // quads start at data+1
        __m128i*        quads(data + 1);

        // scan for each term in sort order
        unsigned int    scan_lo(0);
        unsigned int    scan_hi(size);
        
        bool            match(  bisection_match<(order >> 12) & 0xf>(quad, quads, scan_lo, scan_hi) &&
                                bisection_match<(order >>  8) & 0xf>(quad, quads, scan_lo, scan_hi) &&
                                bisection_match<(order >>  4) & 0xf>(quad, quads, scan_lo, scan_hi) &&
                                bisection_match<(order      ) & 0xf>(quad, quads, scan_lo, scan_hi) );

        // x100 indicates offsets
        if (type & 0x100)
        {
            // scan each block referenced, if the insert suceeded bubble up any blocks added.
        
            // blocks referenced start at data + 202 to allow 201 quads, so can insert one and then split so 
            // less that 200.
            int*    blocks(reinterpret_cast<int*>(data + 202));
            
            for (size_t i(scan_lo); i < scan_hi; ++i) 
            {
                if (insert_b(data, blocks[i], block_added, break_value))
                {
                    if (block_added != 0) // non-zero indicates that a block with that id was added
                    {
                        // move data above i up by one
                        for (size_t index(size); index > i; --index)
                        {
                            blocks[index] = blocks[index - 1];
                            _mm_store_si128(quads + index, _mm_load_si128(quads + index - 1));
                        }
                        
                        // insert reference to new block at i + 1
                        _mm_store_si128(quads + i + 1, break_value);
                        blocks[i + 1] = block_added;
                        
                        if (size < 200) // simple case - increment size and insert reference to new block
                        {
                            type = 0x100 | (size+1);

                            _block_accessor.commit_block(data, block_id);
                            _block_accessor.release_block(data, block_id);
                            
                            block_added = 0;
                        }
                        else // need to split the block
                        {
                            // set size to 101 and write the block back to file
                            type = 0x100 | 100;
                            
                            _block_accessor.commit_block(data, block_id);
                            // _block_accessor.release_block(data, block_id);
                            
                            block_added = _block_accessor.create_block();
                            // rewrite this bit to use get_block to get immutable data, lock_block to get mutable data, 
                            // release_block to release data (commits mutable data changes)
                            // also provide image for switch_block == lock B, copy A to B, relase A, 
                            // which provides a short cut if there's only one ref on A, then no need to copy

                            // __m128i* new_data(_block_accessor.lock_block(block_id));
                            
                            // set size to 100 and shuffle the data down
                            type = 0x100 | 101;
                            
                            for (size_t index(1); index < 101; ++index)
                            {
                                blocks[index] = blocks[index + 100];
                                _mm_store_si128(quads + index, _mm_load_si128(quads + index + 100));
                            }
                            
                            _block_accessor.commit_block(data, block_id);
                            _block_accessor.release_block(data, block_id);
                            
                            block_added = block_id;
                            break_value = _mm_loadu_si128(quads);
                        }
                    }

                    return true;
                }
            }
        }
        else if (!match) // data block - insert if not already present
        {
            if (size < 255)
            {
                // move data above i up by one
                for (size_t index(size); index > scan_hi; --index)
                {
                    _mm_store_si128(quads + index, _mm_load_si128(quads + index - 1));
                }
                        
                // insert reference to new block at i + 1
                _mm_store_si128(quads + scan_hi, quad);
                        
                type = 0x100 | (size+1);

                _block_accessor.commit_block(data, block_id);
                _block_accessor.release_block(data, block_id);
            }
            else // split into two 
            {
                type = 0x100 | 128;

                if (scan_hi > 128)
                {
                    _block_accessor.commit_block(data, block_id);
                    _block_accessor.release_block(data, block_id);
                }
                else
                {
                    for (size_t index(size); index > scan_hi; --index)
                    {
                        _mm_store_si128(quads + index, _mm_load_si128(quads + index - 1));
                    }
                }
//TODO                
            }
        }
    }
    
private:
    block_accessor   _block_accessor;
};


#endif // _B_TREE_H_
